
import FIFO::*;
import ConfigReg::*;
import RWire::*;

import List::*;
import Monad::*;

interface SFIFO#(type alpha_T, type search_T);
   method Action enq(alpha_T x);
   method Action deq();
   method alpha_T first();
   method Action clear();
   method Bool find(search_T x);   
   method Bool find2(search_T x);

endinterface

module mkSFIFO#(function Bool searchfunc(search_T s, alpha_T x)) (SFIFO#(alpha_T, search_T))
    provisos
      (Bits#(alpha_T,asz));

    Reg#(alpha_T) f0 <- mkConfigRegU();
    Reg#(alpha_T) f1 <- mkConfigRegU();

    Reg#(Bool)   vf0 <- mkConfigReg(False);
    Reg#(Bool)   vf1 <- mkConfigReg(False);

    PulseWire edge1      <- mkPulseWire();
      
    method Action enq(alpha_T x) if (!(vf0 && vf1));
      if (edge1 || !vf0)//empty or we're dequeueing
	begin
          vf0 <= True; //True
	  vf1 <= False;
	   f0 <= x;
  	end
      else // !vf1 
        begin
          vf1 <= True;
           f1 <= x;
 	end
    endmethod
										
    method Action deq() if (vf0);
      edge1.send();
      vf0 <= vf1;
       f0 <= f1;
      vf1 <= False;
    endmethod  

    method alpha_T first() if(vf0);
      return (f0);
    endmethod  

    method Action clear();
      vf0 <= False;
      vf1 <= False;
    endmethod

    method Bool find(search_T sv);
      Bool nvf0 = edge1 ? False: vf0;
      Bool nvf1 = vf1;

      return (nvf0 && searchfunc(sv, f0) || 
              nvf1 && searchfunc(sv, f1));
    endmethod

    method Bool find2(search_T sv);
      Bool nvf0 = edge1 ? False: vf0;
      Bool nvf1 = vf1;
      
      return (nvf0 && searchfunc(sv, f0) || 
              nvf1 && searchfunc(sv, f1));
    endmethod

endmodule

module mkSFIFO1#(function Bool searchfunc(search_T s, alpha_T x)) (SFIFO#(alpha_T, search_T))
    provisos
      (Bits#(alpha_T,asz), Eq#(alpha_T));

    Reg#(alpha_T) f0 <- mkConfigRegU;

    Reg#(Bool)   vf0 <- mkConfigReg(False);

    PulseWire edge1      <- mkPulseWire();
      
    method Action enq(alpha_T x) if (!vf0);
      vf0 <= True; //True
      f0 <= x;
    endmethod
										
    method Action deq() if (vf0);
      edge1.send();
      vf0 <= False;
    endmethod  

    method alpha_T first() if(vf0);
      return (f0);
    endmethod  

    method Action clear();
      vf0 <= False;
    endmethod

    method Bool find(search_T sv);
      Bool nvf0 = edge1 ? False: vf0;

      return (nvf0 && searchfunc(sv, f0));
    endmethod

    method Bool find2(search_T sv);
      Bool nvf0 = edge1 ? False: vf0;
      return (nvf0 && searchfunc(sv, f0));
    endmethod

endmodule

